//
//------------------------------------------------------------------------------
// Name    : HPWM2.bas
// Author  : Jerry Messina
// Date    : 25 APR 2023
// Version : 1.00
//
// hardware PWM using CCP1/CCP2/CCP3 and TMR2/4/6
// default options set for CCP2/TMR4
//
// ver 1.00 13 25 APR 2023
//  - initial version, based on original pwm.bas by David John Barker
//  - add support for additional devices, prescalers, and tmr clks  
//  - add MinFreq(), MaxFreq(), Freq(), SetPWM()                    
//  - add #options HPWM_PIN, HPWM_TMR, HPWM_CCP, HPWM_LIMIT, etc    
//  - add user-defined callback event used during setup (PPS, etc)  
//------------------------------------------------------------------------------
//
module HPWM2

//-------------------------------------
// user options
//-------------------------------------

// CCPx selection option (CCP1/2/3, device dependant)  
#option HPWM2_CCP = 2
// check device file '#const _ccp' or '#const _eccp' settings
#if ((_ccp + _eccp) >= 3)       // if supported
  #if not(HPWM2_CCP in (1, 2, 3))
    #error HPWM2_CCP, "invalid HPWM2_CCP option for " + _device
  #endif
#elseif ((_ccp + _eccp) >= 2)
  #if not(HPWM2_CCP in (1, 2))
    #error HPWM2_CCP, "invalid HPWM2_CCP option for " + _device
  #endif
#elseif (_ccp = 1) or (_eccp = 1)
  #if not(HPWM2_CCP in (1))
    #error HPWM2_CCP, "invalid HPWM2_CCP option for " + _device
  #endif
#else
  #error _device + " device file does not support _ccp or _eccp"
#endif      // _ccp or _eccp

// TMRx selection option (TMR2/4/6, device dependant)
#option HPWM2_TMR = 4
// check device file '#const _timer8' setting
#if (_timer8 >= 3)
  #if not(HPWM2_TMR in (2, 4, 6))
    #error HPWM2_TMR, "invalid HPWM2_TMR option for " + _device
  #endif
#elseif (_timer8 = 2)
  #if not(HPWM2_TMR in (2, 4))
    #error HPWM2_TMR, "invalid HPWM2_TMR option for " + _device
  #endif
#elseif (_timer8 = 1)
  #if not(HPWM2_TMR in (2))
    #error HPWM2_TMR, "invalid HPWM2_TMR option for " + _device
  #endif
#else
  #error _device + " device file does not support _timer8"
#endif      // _timer8

// CCP PWM output port pin option
// if you need PPS then set a user-defined event with AssignEvent and handle it there
// on some devices the CCPx output pin can be changed via a CONFIG setting
#if _device in (18F1220, 18F1320)
  #option HPWM2_PIN = PORTB.3
#elseif _device in (18F13K22, 18F14K22, 18LF13K22, 18LF14K22) 
  #option HPWM2_PIN = PORTC.5
#elseif _device in (18F13K50, 18F14K50, 18LF13K50, 18LF14K50) 
  #option HPWM2_PIN = PORTC.5
#endif
#option HPWM2_PIN = PORTC.1           // default CCP2 output pin (typical)
#if not(IsValidPort(HPWM2_PIN) and IsValidPortPin(HPWM2_PIN))
  #error HPWM2_PIN, "invalid HPWM2_PIN selection for " + _device
#endif  
dim PWM_PIO as HPWM2_PIN.HPWM2_PIN@

// default PWM pin output level when stopped
// 0=low, 1=high, 2=set as input (previous version default)
#option HPWM2_PIN_LEVEL = 0
#if not(HPWM2_PIN_LEVEL in (0, 1, 2))
  #error HPWM2_PIN_LEVEL, "invalid HPWM2_PIN_LEVEL setting"
#endif

// limit SetFreq() values to range of MinFreq, MaxFreq
// if option HPWM2_LIMIT_FREQ = true then the pFreq parameter will be adjusted
// to be in the range of MinFreq to MaxFreq, and SetFreq will always return true
#option HPWM2_LIMIT_FREQ = true
#if not(HPWM2_LIMIT_FREQ in (true, false))
  #error HPWM2_LIMIT_FREQ, "invalid HPWM2_LIMIT_FREQ setting"
#endif

// find closest freq
// if option HPWM2_CLOSEST_FREQ = true then SetFreq() will try to find
//  the closest match to the requested pFreq
// otherwise it will use the freq >= requested value (which may be MUCH larger)
// adds about 200 bytes of code, and 100us to the exec time
#option HPWM2_CLOSEST_FREQ = true
#if not(HPWM2_CLOSEST_FREQ in (true, false))
  #error HPWM2_CLOSEST_FREQ, "invalid HPWM2_CLOSEST_FREQ setting"
#endif

// clear user-defined callback event after first use of SetFreq/SetPWM 
// set false if you want the event called every time AssignCCPTMRS() is called
#option HPWM2_CLEAR_EVENT = true
#if not(HPWM2_CLEAR_EVENT in (true, false))
  #error HPWM2_CLEAR_EVENT, "invalid HPWM2_CLEAR_EVENT setting"
#endif

// default autostart setting
// if true then SetFreq() will automatically call Start()
// set false to skip Start... user must call Start after SetFreq (similar to SetPWM)
#option HPWM2_DEFAULT_AUTOSTART = true
#if not(HPWM2_DEFAULT_AUTOSTART in (true, false))
  #error HPWM2_DEFAULT_AUTOSTART, "invalid HPWM2_DEFAULT_AUTOSTART setting"
#endif

// default dutycycle for SetFreq if not specified
#option HPWM2_DEFAULT_DUTYCYCLE = 0
#if not(HPWM2_DEFAULT_DUTYCYCLE in (0 to 100))
  #error HPWM2_DEFAULT_DUTYCYCLE, "invalid HPWM2_DEFAULT_DUTYCYCLE setting"
#endif
const DEFAULT_DUTYCYCLE = HPWM2_DEFAULT_DUTYCYCLE

//-------------------------------------
// end of user options
//-------------------------------------

// define HPWM2_TMR2_TYPE based on device (applies to TMR2/4/6)
//  0 = basic TMR2 (most devices)
//  1 = TMR2 with HLT (K40, Q10, and xv18 devices (K42, K83, all Q))
//      these devices have extended prescalers (1:128) and a programmable clock source
//  2 = basic TMR2 with CCPTMRS0 format change (K22, J13, J53, J94)
//  3 = basic TMR2 with TCLKCON (J11, J50)
//-------------------------------------
// 18FxxQ40/Q41 (_xv18, but requires special handling)
//-------------------------------------
#if _device in (18F04Q40, 18F05Q40, 18F06Q40) or _
    _device in (18F14Q40, 18F15Q40, 18F16Q40) or _
    _device in (18F04Q41, 18F05Q41, 18F06Q41) or _
    _device in (18F14Q41, 18F15Q41, 18F16Q41)
  #define HPWM2_TMR2_TYPE = 1
  dim TMR2 as T2TMR,    // no alias in device file 
      PR2  as T2PR,
      TMR4 as T4TMR,
      PR4  as T4PR
//-------------------------------------
// 18FxxQ71 (_xv18, but requires special handling)
//-------------------------------------
#elseif _device in (18F24Q71, 18F25Q71, 18F26Q71) or _
        _device in (18F44Q71, 18F45Q71, 18F46Q71) or _
        _device in (18F54Q71, 18F55Q71, 18F56Q71)     
  #define HPWM2_TMR2_TYPE = 1
  dim TMR2 as T2TMR,    // no alias in device file 
      PR2  as T2PR,
      TMR4 as T4TMR,
      PR4  as T4PR
//-------------------------------------
// _xv18 devices
//-------------------------------------
#elseif defined(_xv18) and (_xv18 = 1)
  #define HPWM2_TMR2_TYPE = 1
//-------------------------------------
// 18FxxK22 
//-------------------------------------
#elseif _device in (18F23K22, 18F24K22, 18F25K22, 18F26K22, 18LF23K22, 18LF24K22, 18LF25K22, 18LF26K22) or _
        _device in (18F43K22, 18F44K22, 18F45K22, 18F46K22, 18LF43K22, 18LF44K22, 18LF45K22, 18LF46K22) or _
        _device in (18F65K22, 18F66K22, 18F67K22) or _
        _device in (18F85K22, 18F86K22, 18F87K22)
  #define HPWM2_TMR2_TYPE = 2
//-------------------------------------
// 18FxxK40 
//-------------------------------------
#elseif _device in (18F24K40, 18F25K40, 18LF24K40, 18LF25K40) or _
        _device in (18F26K40, 18F27K40, 18LF26K40, 18LF27K40) or _
        _device in (18F45K40, 18F46K40, 18F47K40, 18LF45K40, 18LF46K40, 18LF47K40)
  #define HPWM2_TMR2_TYPE = 1
  dim CCPTMRS0 as CCPTMRS       // alias reg
#elseif _device in (18F65K40, 18F66K40, 18F67K40, 18LF65K40, 18LF66K40, 18LF67K40)
  #define HPWM2_TMR2_TYPE = 1
//-------------------------------------
// 18FxxQ10
//-------------------------------------
#elseif _device in (18F24Q10, 18F25Q10, 18F26Q10, 18F27Q10) or _
        _device in (18F45Q10, 18F46Q10, 18F47Q10)
  #define HPWM2_TMR2_TYPE = 1
  dim CCPTMRS0 as CCPTMRS       // alias reg
//-------------------------------------
// 18FxxJ11
//-------------------------------------
#elseif _device in (18F24J11, 18F25J11, 18F26J11, 18F44J11, 18F45J11, 18F46J11) or _
        _device in (18LF24J11, 18LF25J11, 18LF26J11, 18LF44J11, 18LF45J11, 18LF46J11)
  #define HPWM2_TMR2_TYPE = 3
//-------------------------------------
// 18FxxJ13
//-------------------------------------
#elseif _device in (18F26J13, 18F27J13, 18F46J13, 18F47J13) or _
        _device in (18LF26J13, 18LF27J13, 18LF46J13, 18LF47J13)
  #define HPWM2_TMR2_TYPE = 2
//-------------------------------------
// 18FxxJ50
//-------------------------------------
#elseif _device in (18F24J50, 18F25J50, 18F26J50, 18F44J50, 18F45J50, 18F46J50) or _
        _device in (18LF24J50, 18LF25J50, 18LF26J50, 18LF44J50, 18LF45J50, 18LF46J50)
  #define HPWM2_TMR2_TYPE = 3
//-------------------------------------
// 18FxxJ53
//-------------------------------------
#elseif _device in (18F26J53, 18F27J53, 18F46J53, 18F47J53) or _
        _device in (18LF26J53, 18LF27J53, 18LF46J53, 18LF47J53)
  #define HPWM2_TMR2_TYPE = 2
//-------------------------------------
// 18FxxJ94/J99
//-------------------------------------
#elseif _device in (18F65J94, 18F66J94, 18F67J94) or _
        _device in (18F85J94, 18F86J94, 18F87J94) or _
        _device in (18F95J94, 18F96J94, 18F97J94) or _
        _device in (18F66J99, 18F86J99, 18F96J99)
  #define HPWM2_TMR2_TYPE = 2
//-------------------------------------
// all other devices
//-------------------------------------
#else
  #define HPWM2_TMR2_TYPE = 0
#endif

// TMRx register aliases based on TMR option
#if (HPWM2_TMR = 2)
dim TxCON as T2CON,
    TMRx  as TMR2,
    PRx   as PR2
#elseif (HPWM2_TMR = 4)
dim TxCON as T4CON,
    TMRx  as TMR4,
    PRx   as PR4
#elseif (HPWM2_TMR = 6)
dim TxCON as T6CON,
    TMRx  as TMR6,
    PRx   as PR6
#endif

// CCPx register aliases based on CCP options
#if (HPWM2_CCP = 1)
dim CCPxCON as CCP1CON,
    CCPRxL  as CCPR1L,
    CCPRxH  as CCPR1H
#elseif (HPWM2_CCP = 2)
dim CCPxCON as CCP2CON,
    CCPRxL  as CCPR2L,
    CCPRxH  as CCPR2H
#elseif (HPWM2_CCP = 3)         // only available on some devices (K22, J13, J53)
dim CCPxCON as CCP3CON,
    CCPRxL  as CCPR3L,
    CCPRxH  as CCPR3H
#endif

// programmable TMR clock source register aliases
// K40, Q10, and xv18 devices (K42, K83, all Q)
#if (HPWM2_TMR2_TYPE = 1)
  #if (HPWM2_TMR = 2)
    dim TxCLKCON as T2CLKCON
  #elseif (HPWM2_TMR = 4)
    dim TxCLKCON as T4CLKCON
  #elseif (HPWM2_TMR = 6)
    dim TxCLKCON as T6CLKCON
  #endif
#endif

// device dependant registers/bits based on TMR2 type
#if (HPWM2_TMR2_TYPE = 0) or (HPWM2_TMR2_TYPE = 2) or (HPWM2_TMR2_TYPE = 3)
  const MAX_PRESCALER = 16,     // supports prescalers 1, 4, 16 
        PRESCALER_MULT  = 4,    // x4 increment
        PRESCALER_SHIFT = 2,    // x4 increment
        PRESCALER_MASK = $03,   // TxCON TxCKPS<1:0> bits
        TMR_ON = 2              // TxCON.TMRxON bit 2
  const                         // CCPxCON bits
        CCP_MODE_MASK as byte = $F0,    // retain DCxB[1:0] duty cycle bite
        CCP_MODE_PWM  as byte = $0C,
        CCPCON_INIT   as byte = $00     // no addtl bits to set
#elseif (HPWM2_TMR2_TYPE = 1)
  const MAX_PRESCALER = 128,    // supports prescalers 1, 2, 4, 8, 16, 32, 64, 128
        PRESCALER_MULT  = 2,    // x2 increment
        PRESCALER_SHIFT = 1,    // x2 increment
        PRESCALER_MASK = $70,   // TxCON CKPS<2:0> bits
        TMR_ON = 7              // TxCON.TxON bit 7
  const                         // CCPxCON bits
        CCP_MODE_MASK as byte = $E0,    // retain EN, clear FMT bits
        CCP_MODE_PWM  as byte = $0C,
        CCPCON_INIT   as byte = $80     // set EN = 1
#else
  #error HPWM2_TMR2_TYPE, "unsupported HPWM2_TMR2_TYPE"
#endif

// variables and constants
const FOSC as longword = _clock * 1000000   // system clock, in Hz
    
dim FMaxDuty as word = 0    // max duty cycle value, 0-1023 (10-bits)

// mode flags
structure flags_t
    b as byte
    autostart as b.booleans(0)      // automatically call Start in SetFreq()
end structure
dim flags as flags_t

// PWM CCP TMR and clock source selection
// --------------------------------------
// PWM defaults to using CCP1 and TMR2
// if you change this then you must assign the TMRx resource to the CCPx peripheral
//  how this is done varies from device to device, but typically it's done using the 
//  CCPTMRS, CCPTMRS0, or TCLKCON register.
// the AssignCCPTMRS() routine can determine how to setup some devices, but unfortunately 
//  there are too many possible permutations, so here we define an event that is called by
//  AssignCCPTMRS() prior to starting the PWM that can be used to setup any required registers.
// This can also be used to init any other hdw settings you might need to change, such
//  as the pin analog/digital mode, PPS assignments, etc.
// NOTE: this module assumes that the TMR is clocked by FOSC/4 (instruction cycle)
//  if your device has a programmable TMR clock source it must be set to match (TxCLK/TxCLKCON)
//
// use SetEvent() to set your user-defined event prior to calling SetFreq or SetPWM
//  if HPWM2_CLEAR_EVENT is true then the user event will only be called the first time
//  AssignCCPTMRS() is used
//
// example usage:
// ' in you main program create an event with whatever code is required
//    event user_hpwm_init()
//        CCPTMRS = 2       // device-dependant code
//        ANSELB.0 = 0      // set pin to digital mode
//    end event
//
// ' assign the event prior to calling SetFreq or SetPWM the first time   
//    hpwm.SetEvent(user_hpwm_init)
//    hpwm.SetFreq(5000)
//
type event_t = event()
dim AssignCCPTMRSEvent as event_t = 0

public inline sub SetEvent(user_event as event_t)
    AssignCCPTMRSEvent = user_event
end sub

// assign CCP TMR selection and call user 'AssignCCPTMRSEvent'
private sub AssignCCPTMRS()
  // we know how to set some devices... user must handle others
  #if (HPWM2_TMR2_TYPE = 1)                  // K40, Q10, all _xv18
    // set TMR clock source
    TxCLKCON = %0001                            // TMRx clock = FOSC/4
    // assign ccp tmr
    #if (HPWM2_CCP = 1) and (HPWM2_TMR = 2)       // CCPTMRS0.C1TSEL
      CCPTMRS0.bits(1) = 0
      CCPTMRS0.bits(0) = 1
    #elseif (HPWM2_CCP = 1) and (HPWM2_TMR = 4)
      CCPTMRS0.bits(1) = 1
      CCPTMRS0.bits(0) = 0
    #elseif (HPWM2_CCP = 1) and (HPWM2_TMR = 6)
      CCPTMRS0.bits(1) = 1
      CCPTMRS0.bits(0) = 1
    #elseif (HPWM2_CCP = 2) and (HPWM2_TMR = 2)   // CCPTMRS0.C2TSEL
      CCPTMRS0.bits(3) = 0
      CCPTMRS0.bits(2) = 1
    #elseif (HPWM2_CCP = 2) and (HPWM2_TMR = 4)
      CCPTMRS0.bits(3) = 1
      CCPTMRS0.bits(2) = 0
    #elseif (HPWM2_CCP = 2) and (HPWM2_TMR = 6)
      CCPTMRS0.bits(3) = 1
      CCPTMRS0.bits(2) = 1
    #elseif (HPWM2_CCP = 3) and (HPWM2_TMR = 2)   // CCPTMRS0.C3TSEL
      CCPTMRS0.bits(5) = 0
      CCPTMRS0.bits(4) = 1
    #elseif (HPWM2_CCP = 3) and (HPWM2_TMR = 4)
      CCPTMRS0.bits(5) = 1
      CCPTMRS0.bits(4) = 0
    #elseif (HPWM2_CCP = 3) and (HPWM2_TMR = 6)
      CCPTMRS0.bits(5) = 1
      CCPTMRS0.bits(4) = 1
    #endif

  #elseif (HPWM2_TMR2_TYPE = 2)              // K22, J13, J53, J94
    #if (HPWM2_CCP = 1) and (HPWM2_TMR = 2)       // CCPTMRS0.C1TSEL
      CCPTMRS0.bits(2) = 0                      // (K22 unimplemented bit 2)
      CCPTMRS0.bits(1) = 0
      CCPTMRS0.bits(0) = 0
    #elseif (HPWM2_CCP = 1) and (HPWM2_TMR = 4)
      CCPTMRS0.bits(2) = 0
      CCPTMRS0.bits(1) = 0
      CCPTMRS0.bits(0) = 1
    #elseif (HPWM2_CCP = 1) and (HPWM2_TMR = 6)
      CCPTMRS0.bits(2) = 0
      CCPTMRS0.bits(1) = 1
      CCPTMRS0.bits(0) = 0
    #elseif (HPWM2_CCP = 2) and (HPWM2_TMR = 2)   // CCPTMRS0.C2TSEL
      CCPTMRS0.bits(5) = 0
      CCPTMRS0.bits(4) = 0
      CCPTMRS0.bits(3) = 0
    #elseif (HPWM2_CCP = 2) and (HPWM2_TMR = 4)
      CCPTMRS0.bits(5) = 0
      CCPTMRS0.bits(4) = 0
      CCPTMRS0.bits(3) = 1
    #elseif (HPWM2_CCP = 2) and (HPWM2_TMR = 6)
      CCPTMRS0.bits(5) = 0
      CCPTMRS0.bits(4) = 1
      CCPTMRS0.bits(3) = 0
    #elseif (HPWM2_CCP = 3) and (HPWM2_TMR = 2)   // CCPTMRS0.C3TSEL
      CCPTMRS0.bits(7) = 0
      CCPTMRS0.bits(6) = 0
    #elseif (HPWM2_CCP = 3) and (HPWM2_TMR = 4)
      CCPTMRS0.bits(7) = 0
      CCPTMRS0.bits(6) = 1
    #elseif (HPWM2_CCP = 3) and (HPWM2_TMR = 6)
      CCPTMRS0.bits(7) = 1
      CCPTMRS0.bits(6) = 0
    #endif

  #elseif (HPWM2_TMR2_TYPE = 3)                  // J11, J50
    // you may have to adjust these in the AssignCCPTMRSEvent
    // if using both CCP1 and CCP2 since they're tied together
    //  10 = ECCP1 and ECCP2 both use Timer4 (PWM)
    //  01 = ECCP1 uses Timer2 (PWM); ECCP2 uses Timer4 (PWM)
    //  00 = ECCP1 and ECCP2 both use Timer2 (PWM)
    #if (HPWM2_CCP = 1) and (HPWM2_TMR = 2)       // TCLKCON.T3CCP1
      TCLKCON.bits(1) = 0
      TCLKCON.bits(0) = 0
    #elseif (HPWM2_CCP = 1) and (HPWM2_TMR = 4)
      TCLKCON.bits(1) = 1
      TCLKCON.bits(0) = 0
    #elseif (HPWM2_CCP = 2) and (HPWM2_TMR = 2)
      TCLKCON.bits(1) = 0
      TCLKCON.bits(0) = 0
    #elseif (HPWM2_CCP = 2) and (HPWM2_TMR = 4)
      TCLKCON.bits(1) = 1
      TCLKCON.bits(0) = 0
    #endif

  #else     // HPWM2_TMR2_TYPE = 0
    // check for a change in the default usage of CCP1 and TMR2 
    // and warn the user to assign an event to handle it
    #if (HPWM2_TMR <> 2)
      #warning "you must assign an event to handle changing the default CCP/TMR2 settings"
    #endif
  #endif    // HPWM2_TMR2_TYPE
  
    // call user-defined event to complete any setup
    AssignCCPTMRSEvent()

    // clear event after first use (if enabled)
  #if (HPWM2_CLEAR_EVENT)
    AssignCCPTMRSEvent = 0
  #endif
end sub

// set TMR prescaler
//  parameter: prescaler = prescaler divider (ie 1/4/16, etc)
// returns: true=prescaler valid, false=prescaler invalid
public function SetTMRPrescaler(prescaler as byte) as boolean
    result = true
    TxCON.bits(TMR_ON) = 0                      // turn off timer   
  #if (HPWM2_TMR2_TYPE = 0) or (HPWM2_TMR2_TYPE = 2) or (HPWM2_TMR2_TYPE = 3) // prescalers 1, 4, 16 
    select (prescaler)                          // configure TxCON prescaler
        case 1:   prescaler = %00000000            // prescale 1
        case 4:   prescaler = %00000001            // prescale 4
        case 16:  prescaler = %00000010            // prescale 16  (typ '1x', but see GetTMRPrescaler)
        else
            prescaler = 0
            result = false
    end select

  #elseif (HPWM2_TMR2_TYPE = 1)   // prescalers 1, 2, 4, 8, 16, 32, 64, 128
    select (prescaler)                          // configure TxCON prescaler
        case 1:   prescaler = %00000000            // prescale 1
        case 2:   prescaler = %00010000            // prescale 2
        case 4:   prescaler = %00100000            // prescale 4
        case 8:   prescaler = %00110000            // prescale 8
        case 16:  prescaler = %01000000            // prescale 16
        case 32:  prescaler = %01010000            // prescale 32
        case 64:  prescaler = %01100000            // prescale 64
        case 128: prescaler = %01110000            // prescale 128
        else
            prescaler = 0
            result = false
    end select
  #endif
    // set prescaler reg bits
    TxCON = (TxCON and not(PRESCALER_MASK)) or prescaler
end function

// get TMR prescaler
//  returns prescaler divider based on TxCON register setting
public function GetTMRPrescaler() as byte
    dim prescaler as byte
    
    prescaler = TxCON and PRESCALER_MASK
  #if (HPWM2_TMR2_TYPE = 0) or (HPWM2_TMR2_TYPE = 2) or (HPWM2_TMR2_TYPE = 3) // prescalers 1, 4, 16 
    select (prescaler)
        case %00: prescaler = 1
        case %01: prescaler = 4
        case %10: prescaler = 16            // some devices specifically call out '10' (47J13)
        case %11: prescaler = 16            // but most devices say '1x'
    end select

  #elseif (HPWM2_TMR2_TYPE = 1)   // prescalers 1, 2, 4, 8, 16, 32, 64, 128
    select (prescaler)
        case %00000000: prescaler = 1
        case %00010000: prescaler = 2
        case %00100000: prescaler = 4
        case %00110000: prescaler = 8
        case %01000000: prescaler = 16
        case %01010000: prescaler = 32
        case %01100000: prescaler = 64
        case %01110000: prescaler = 128
    end select
  #endif
    result = prescaler
end function

// given a prescaler value (1/4/16 or 1/2/4/...128) get the next prescaler
//   returns 0 when prescaler is max
public function GetNextPrescaler(prescaler as byte) as byte
    result = 0
    if (prescaler < MAX_PRESCALER) then
        result = prescaler * PRESCALER_MULT    // next prescaler value
    endif
end function

// set duty cycle, 0-1023 (10-bits)
//  HPWM2_TMR2_TYPE = 0, 2, or 3:
//    The CCPRxL contains the eight msbs and CCPxCON<5:4> contains the two lsbs
//    The 10-bit value is represented by CCPRxL:CCPxCON<5:4>
//  HPWM2_TMR2_TYPE = 1 (with CCPxCON.FMT=0, right-aligned format):
//    The CCPRxL contains the eight lsbs and CCPRxH<1:0> contains the two msbs
//    The 10-bit value is represented by CCPRxH<1:0>::CCPRxL
public sub SetDuty(pDuty as word)
    // limit duty cycle setting to 1023 (max 10-bit setting)
    if (pDuty > 1023) then
        pDuty = 1023
    endif        
  #if (HPWM2_TMR2_TYPE = 0) or (HPWM2_TMR2_TYPE = 2) or (HPWM2_TMR2_TYPE = 3)
    CCPxCON.bits(5) = pDuty.1       // 2 lsb's go to DCxB[1:0]
    CCPxCON.bits(4) = pDuty.0  
    CCPRxL = pDuty >> 2
  #elseif (HPWM2_TMR2_TYPE = 1)      // with CCPxCON.FMT=0 (right-aligned format)
    CCPRxL = pDuty.bytes(0)         // lsb
    CCPRxH.bits(0) = pDuty.bits(8)  // msb's
    CCPRxH.bits(1) = pDuty.bits(9)
  #endif
end sub

// get duty cycle
//  returns 10-bit duty cycle value, 0-1023
public function GetDuty() as word
  #if (HPWM2_TMR2_TYPE = 0) or (HPWM2_TMR2_TYPE = 2) or (HPWM2_TMR2_TYPE = 3)
    result = CCPRxL << 2
    result.bits(1) = CCPxCON.bits(5)
    result.bits(0) = CCPxCON.bits(4)
  #elseif (HPWM2_TMR2_TYPE = 1)      // with CCPxCON.FMT=0
    result = CCPRxL
    result.bits(8) = CCPRxH.bits(0)
    result.bits(9) = CCPRxH.bits(1)
  #endif
end function

// set FMaxDuty based on PRx period value
public sub SetMaxDuty(PRval as byte)
    FMaxDuty = (PRval + 1) * 4
    if (FMaxDuty > 1023) then
        FMaxDuty = 1023             // 10-bit limit
    endif
end sub

// return calculated MaxDuty cycle
public inline function MaxDuty() as FMaxDuty
end function

// given a percentage 0-100, return duty cycle word
public function GetDutyPercent(pPercent as byte) as word
    dim dw as longword

    if (pPercent > 100) then
        pPercent = 100
    endif
    // result = ((MaxDuty() * pPercent) + 50) / 100
    dw = MaxDuty() * pPercent
    dw = dw + 50                   // round up
    result = dw/100
end function

// set the duty cycle as a percentage, 0 to 100
public sub SetDutyPercent(pPercent as byte)
    SetDuty(GetDutyPercent(pPercent))
end sub

// get current PRx period register
public inline function GetPR() as PRx
end function

// start PWM
public sub Start()
    CCPxCON = (CCPxCON and CCP_MODE_MASK) or (CCP_MODE_PWM or CCPCON_INIT)
    output(PWM_PIO)        // enable PWM out on the CCPx pin 
    TxCON.bits(TMR_ON) = 1
end sub

// stop PWM
public sub Stop()
  #if (HPWM2_PIN_LEVEL = 0)  // set PWM pin output level
    low(PWM_PIO)
  #elseif (HPWM2_PIN_LEVEL = 1)
    high(PWM_PIO)
  #else                     // set PWM pin to input (original version)
    input(PWM_PIO)
  #endif
    CCPxCON = $00           // clearing CCPxCON will relinquish control of the CCPx pin
end sub

// set pwm period, prescaler, and duty cycle registers (low-level)
// parameters:
//   PR val      byte, PRx period register setting
//   prescaler   byte, prescaler setting - either 1/4/16 or 1/2/4/8/16/32/64/128 (if supported)
//   duty cycle  word, 0-1023
// user must call Start() to begin PWM
// exec: 5-6us @ 64MHz
public sub SetPWM(PRval as byte, prescaler as byte=1, duty as word=0)
    TxCON.bits(TMR_ON) = 0          // turn off timer   
    SetTMRPrescaler(prescaler)      // configure TxCON prescaler
    PRx = PRval                     // set period register
    TMRx = PRx                      // and count = period
    AssignCCPTMRS()                 // set CCP TMR selection and call user-defined event
    SetMaxDuty(PRx)                 // maximum duty cycle resolution based on PR2 value  
    SetDuty(duty)                   // set duty cycle
end sub

//
// SetFreq and related functions MaxFreq(), MinFreq(), and Freq()
//
// get max pwm freq
public inline function MaxFreq() as longword
    result = FOSC/ ((0+1)*4*1)          // min value for PR2 is 0
end function

// get min pwm freq
public inline function MinFreq() as longword
    result = FOSC/ ((255+1)*4*MAX_PRESCALER)
end function

// given PR period (0-255) and prescaler values (1/4/16, etc) compute freq
public function Freq(PRval as byte, prescaler as byte) as longword
    dim wprescaler as word           // must be word for values >= 64 and x4

    // result = FOSC/ ((PRx + 1) * (prescaler*4))
    wprescaler.bytes(0) = prescaler
    wprescaler = wprescaler.bytes(0) * 4          // now it's a word, x4
    result = FOSC/ ((PRval + 1) * wprescaler)
end function

// get current actual pwm freq
public function Freq() as longword
    result = Freq(PRx, GetTMRPrescaler())
end function

// set flags
public sub Autostart(b as boolean)
    flags.autostart = b
end sub

// set pwm freq and duty cycle
//  parameters:
//    pFreq longword  desired PWM freq, in Hz
//    pDuty byte      duty cycle in percent, 0-100
//  returns: true=pFreq within range, false=pFreq not within range (see HPWM2_LIMIT_FREQ)
// the actual PWM freq depends on option HPWM2_CLOSEST_FREQ
//  with HPWM2_CLOSEST_FREQ=false the output will be the next freq >= pFreq,
//  which can be a big difference!
//  setting HPWM2_CLOSEST_FREQ=true will find the closest freq to the requested pFreq
// if flags.autostart = true then Start() is called to begin PWM
//  otherwise user must call Start()
// exec: 145-195us @ 64MHz, HPWM2_CLOSEST_FREQ=true adds about 100us
public function SetFreq(pFreq as longword, pDuty as byte=DEFAULT_DUTYCYCLE) as boolean
    dim prescaler as byte
    dim PRval, period as word
    dim wDuty as word
    
  #if (HPWM2_CLOSEST_FREQ)
    dim df, dfn as longint
  #endif    // HPWM2_CLOSEST_FREQ
    
  #if (HPWM2_LIMIT_FREQ)                 // range check and adjust if necessary
    if (pFreq < MinFreq()) then
        pFreq = MinFreq()
    elseif (pFreq > MaxFreq()) then
        pFreq = MaxFreq()
    endif                
  #endif    // HPWM2_LIMIT_FREQ
  
    // compute period range
    // this should always produce a result >= desired freq
    period = FOSC /4 /pFreq
    
    // loop through all the valid prescalers...
    prescaler = 1
    result = false
    repeat
        PRval = (period / prescaler) - 1            // calculate a PR2 value
        if (PRval < 256) then                       // if it's a valid value, then...
            result = true                           // success
            // if enabled, check to see if there's a closer freq match
          #if (HPWM2_CLOSEST_FREQ)
            if (PRval > 0) and (PRval < 255) then   // not at end of a prescaler range 
                df = Freq(PRval, prescaler)         // get actual freq for this PRval
                df = df - pFreq                     // and calc delta from desired
                if (df > 0) then                    // actual f > pFreq... look at next lower
                    dfn = Freq(PRval+1, prescaler)
                    dfn = dfn - pFreq               // calc delta for lower freq
                    if (dfn < 0) then               // abs(dfn)
                        dfn = -dfn
                    endif                    
                    if (dfn < df) then              // lower freq is closer... use that
                        PRval = PRval+1
                    endif
                endif                
            endif
          #endif    // HPWM2_CLOSEST_FREQ
            // set pwm values
            SetMaxDuty(PRval)                       // max duty cycle is based on PRx value  
            wDuty = GetDutyPercent(pDuty)           // convert percent to setting
            SetPWM(PRval, prescaler, wDuty)
            if (flags.autostart) then               // if enabled...
                Start()                             // start PWM
            endif
            exit
        endif   
        prescaler = GetNextPrescaler(prescaler)     // get next prescaler (0=end of list)
    until (prescaler = 0)
end function

// module init
flags.b = 0
#if (HPWM2_DEFAULT_AUTOSTART)
flags.autostart = true
#endif

end module

